################################################################
# File -- Compute_POR_Elasticities.R
# This file contains code used to compute features of the distribution of household elasticities in
# "A Discrete Choice Model for Partially Ordered Alternatives" by Eleni Aristodemou and Adam Rosen.  
# The data used in the paper is from Kantar FMCG purchase panel and is not publicly available.  
#################################################################

rm(list=ls())

# Set working directory

dtapath <- "Estimation_Data_Table4_Aristodemou_and_Rosen(2022).dta"

if (!require("abind")) install.packages("abind")                         
library(abind) 
if (!require("mvtnorm")) install.packages("mvtnorm")                         
library(mvtnorm)    
if (!require("pbivnorm")) install.packages("pbivnorm")  
library(pbivnorm)
if (!require("lattice")) install.packages("lattice")
library(lattice)
if (!require("Rsolnp")) install.packages("Rsolnp")  
library(Rsolnp)
if (!require("matrixcalc")) install.packages("matrixcalc")  
library(matrixcalc)
if (!require("foreign")) install.packages("foreign")
library(foreign)
if (!require("readxl")) install.packages("readxl")
library(readxl)
if (!require("haven")) install.packages("haven")
library(haven)

source("POR_Load_Data_Functions.R")
source("POR_Likelihood_Functions_Final.R")

rseed <- 987698764
set.seed(rseed)

dim_theta <- 16
theta <- c(1.00,0.80,-1.50,-1.20,-0.20,-1.00,-0.10,0.30,0.70,.20,-1.00,-0.10,0.30,0.70,0.50,1.00)

lower = rep(-6,dim_theta)
upper = rep(6,dim_theta)
lower[(dim_theta-1)] <- -1.0
upper[(dim_theta-1)] <- 1.0
lower[dim_theta] <- 0.001
upper[dim_theta] <- 4
stepsize <- rep(0.2,dim_theta)
stepsize[dim_theta-1] <- 0.01
stepsize[dim_theta] <- 0.05
alpha <- 0.05

# LOAD DATA HERE
hhdata <- Load_POR_Data_IFS_16par(dtapath=dtapath)

# Set the number of randomly drawn starting values for gosolnp optimization searches here.
gosolnp_fullML_simulations <- 10000
gosolnp_profileML_simulations <- 500
gosolnp_search_simulations <- 200

k.full.likelihood <- function(theta_minus_k_arg, f_data, theta_k_val, kpos) {
  theta_arg <- append(theta_minus_k_arg, theta_k_val, after=kpos)
  return(neg.full.likelihood(theta = theta_arg, f_data = f_data))
}

k.profile.likelihood <- function(theta_k_val, theta_minus_k_start, LB_minus_k, UB_minus_k, f_data, k, nsim) {
  gosolnp_result <- gosolnp(par = theta_minus_k_start, fun = k.full.likelihood, LB = LB_minus_k, UB = UB_minus_k,
                            control = list(trace = 0), n.sim = nsim, rseed = rseed,
                            f_data = f_data, theta_k_val = theta_k_val, kpos = (k-1))
  solnp_result <- solnp(par = theta_minus_k_start, fun = k.full.likelihood, LB = LB_minus_k, UB = UB_minus_k,
                        control = list(trace = 0), f_data = f_data, theta_k_val = theta_k_val, kpos = (k-1))
  if (gosolnp_result$values[length(gosolnp_result$values)] <= solnp_result$values[length(solnp_result$values)]) {
    return(gosolnp_result)
  } else {
    return(solnp_result)
  }
}

profileLL_values_theta0 <- rep(0,length(theta))
LRstat <- rep(Inf,length(theta))
CI_ptID_lower_bounds <- rep(-Inf,length(theta)) # Confidence interval lower bounds using conventional ML asymptotic normality.
CI_ptID_upper_bounds <- rep(Inf,length(theta)) # Confidence interval upper bounds using conventional ML asymptotic normality.
CI_CCT_M3_lower_bounds <- rep(-Inf,length(theta)) # Confidence interval lower bounds using Chen, Christensen, Tamer (2017) method 3.
CI_CCT_M3_upper_bounds <- rep(Inf,length(theta)) # Confidence interval upper bounds using Chen, Christensen, Tamer (2017) method 3.
Ln_max <- -Inf
theta_hat <- rep(0,length(theta))
std_errs <- rep(Inf,length(theta))

c_alpha <- qchisq(1-alpha,1) # critical value
tol <- 0.00001

# SEARCH FOR THE MAXIMIZER
  
  search.time.solnp <- system.time(
    solnp_LL_full <- solnp(par = theta, fun = neg.full.likelihood, LB = lower, UB = upper, control = list(trace = 0),
                           f_data = hhdata))
  cat("Full likelihood maximimzation by solnp took ", search.time.solnp[3], " seconds.\n", sep="")
  cat("Maximum value of the log likelihood was ", -solnp_LL_full$values[length(solnp_LL_full$values)], "\n", sep ="")
  
  search.time.gosolnp <- system.time(
    gosolnp_LL_full <- gosolnp(par = theta, fun = neg.full.likelihood, LB=lower, UB=upper, control = list(trace = 0),
                             n.sim = gosolnp_fullML_simulations, rseed = rseed, f_data = hhdata))
  cat("Full likelihood maximimzation by gosolnp using n.sim = ", gosolnp_fullML_simulations, " took ", search.time.gosolnp[3], " seconds.\n", sep="")
  cat("Maximum value of the log likelihood was ", -gosolnp_LL_full$values[length(gosolnp_LL_full$values)], "\n\n", sep ="")
  
  if (gosolnp_LL_full$values[length(gosolnp_LL_full$values)] < solnp_LL_full$values[length(solnp_LL_full$values)]) {
    LL_full_obj <- gosolnp_LL_full
  } else {
    LL_full_obj <- solnp_LL_full
  }
  Ln_max <- -LL_full_obj$values[length(LL_full_obj$values)]
  theta_hat <- LL_full_obj$pars
  hessian <- LL_full_obj$hessian
  
  if (is.positive.definite(hessian)) (std_errs <- sqrt(diag(chol2inv(chol(hessian)))))
  CI_ptID_lower_bounds <- theta_hat - qnorm(1-alpha/2) * std_errs
  CI_ptID_upper_bounds <- theta_hat + qnorm(1-alpha/2) * std_errs
  
  cat("Theta hat = (")
  cat(theta_hat, sep=", ")
  cat(")\n")
  
  cat("Lower CI endpoints = (")
  cat(CI_ptID_lower_bounds, sep=", ")
  cat(")\n")
  
  cat("Upper CI endpoints = (")
  cat(CI_ptID_upper_bounds, sep=", ")
  cat(")\n")
  
  elasticities <- Compute_price_elasticities(theta=theta_hat,hhdata$X1,hhdata$X2,hhdata$P)
  
  sink("elasticity_output.txt", split=T)
  
  cat("---------- MEAN ELASTICITIES OF THE OUTSIDE ALTERNATIVE WITH RESPECT TO PRODUCT PRICES ---------- \n")
  for (i in 1:2) {
    for (j in 1:2 ) {
      cat("Estimated mean elasticity of outside alternative wrt price of product (", i,",",j, "):  ", mean(elasticities$outside_elasticities[,i,j]),"\n", sep ="")
    }
  }
  cat("\n")
  
  cat("---------- MEAN ELASTICITIES OF EACH PRODUCT WITH RESPECT TO PRICES ---------- \n")
  for (i in 1:2) {
    for (j in 1:2 ) {
      for (k in 1:2) {
        for (l in 1:2) {
          cat("Estimated mean elasticity of product, (",i,",",j,") wrt price of product (", k,",",l, "):  ", mean(elasticities$POR_elasticities[,i,j,k,l]),"\n", sep ="")        
        }
      }
    }
  }
  
  cat("\n")
  
  cat("---------- QUANTILES OF ELASTICITIES OF THE OUTSIDE ALTERNATIVE WITH RESPECT TO PRODUCT PRICES ---------- \n")
  for (i in 1:2) {
    for (j in 1:2 ) {
      cat("Estimated 0.2 quantile of elasticity of outside alternative wrt price of product(", i,",",j, "): ", quantile(elasticities$outside_elasticities[,i,j], probs=0.2),"\n",sep="")
      cat("Estimated median elasticity of outside alternative wrt price of product (", i,",",j, "):  ", median(elasticities$outside_elasticities[,i,j]),"\n", sep ="")
      cat("estimated 0.8 quantile of elasticity of outside alternative wrt to price of product(", i,",",j, ": ", quantile(elasticities$outside_elasticities[,i,j], probs=0.8),"\n",sep="")
    }
  }
  cat("\n")
  
  cat("---------- QUANTILES OF ELASTICITIES OF EACH PRODUCT WITH RESPECT TO PRICES ------------- \n")
  for (i in 1:2) {
    for (j in 1:2) {
      for (k in 1:2) {
        for (l in 1:2) {
          cat("Estimated 0.2 quantile of elasticity of product (",i,",",j,") wrt price of product (", k,",",l, "): ", quantile(elasticities$POR_elasticities[,i,j,k,l], probs=0.2),"\n", sep="")
          cat("Estimated median elasticity of product (",i,",",j,") wrt price of product (", k,",",l, "):  ", median(elasticities$POR_elasticities[,i,j,k,l]),"\n", sep ="")        
          cat("Estimated 0.8 quantile of elasticity of product (",i,",",j,") wrt price of product (", k,",",l, "): ", quantile(elasticities$POR_elasticities[,i,j,k,l], probs=0.8), "\n", sep="")
        }
      }
    }
  }
  
sink()  
  