# This file contains functions for computing confidence sets for individual parameter components

rm(list=ls())

# Set working directory and data path as needed

dtapath <-"Estimation_Data_Table4_Aristodemou_and_Rosen(2022).dta" # To produce results in Table 4
#dtapath <- "Estimation_Data_Table5_Aristodemou_and_Rosen(2022).dta" # To produce results in Table 5

#install.packages("mvtnorm") # uncomment if package needs to be installed.
library("mvtnorm")
#install.packages("pbivnorm") # uncomment if package needs to be installed.
library(pbivnorm)
#install.packages("Rsolnp") # uncomment if package needs to be installed.
library(Rsolnp)
#install.packages("lattice") # uncomment if package needs to be installed.
library(lattice)
#install.packages("matrixcalc") # uncomment if package needs to be installed.
library(matrixcalc)
library(foreign)
library(readxl)
library(haven)

source("POR_Load_Data_Functions.R")
source("POR_Likelihood_Functions_Final.R")

rseed <- 987698764
set.seed(rseed)

theta <- c(1.00,0.80,-1.50,-1.20,-0.20,-1.00,-0.10,0.30,0.70,.20,-1.00,-0.10,0.30,0.70,0.50,1.00)
# starting value for optimizer

lower = rep(-6,16)
upper = rep(6,16)
lower[15] <- -1.0
upper[15] <- 1.0
lower[16] <- 0.001
upper[16] <- 4
stepsize <- rep(0.2,16)
stepsize[15] <- 0.01
stepsize[16] <- 0.05
alpha <- 0.05

# LOAD DATA HERE
hhdata <- Load_POR_Data_IFS_16par(dtapath=dtapath)

# Set the number of randomly drawn starting values for gosolnp optimization searches here.
gosolnp_fullML_simulations <- 10000
gosolnp_profileML_simulations <- 500
gosolnp_search_simulations <- 200

######### Set parameter indices for which to compute QLR confidence sets here ######################
###### e.g. parameter_indices <- seq(5,8,1) for the 5th through 8th components of the parameter vector theta ######
###### e.g. parameter_indices <- seq(k,h,1) for the k-th through h-th components of the parameter vector theta, etc. ######
###### For results in the paper the code in this file was executed separately with one of the four lines below uncommented. They may be run on separate nodes simultaneously.
parameter_indices <- seq(1,4,1)
# parameter_indices <- seq(5,8,1)
# parameter_indices <- seq(9,12,1)
# parameter_indices <- seq(13,16,1)
#############################################

k.full.likelihood <- function(theta_minus_k_arg, f_data, theta_k_val, kpos) {
  theta_arg <- append(theta_minus_k_arg, theta_k_val, after=kpos)
  return(neg.full.likelihood(theta = theta_arg, f_data = f_data))
}

k.profile.likelihood <- function(theta_k_val, theta_minus_k_start, LB_minus_k, UB_minus_k, f_data, k, nsim) {
  gosolnp_result <- gosolnp(par = theta_minus_k_start, fun = k.full.likelihood, LB = LB_minus_k, UB = UB_minus_k,
                            control = list(trace = 0), n.sim = nsim, rseed = rseed,
                            f_data = f_data, theta_k_val = theta_k_val, kpos = (k-1))
  solnp_result <- solnp(par = theta_minus_k_start, fun = k.full.likelihood, LB = LB_minus_k, UB = UB_minus_k,
                        control = list(trace = 0), f_data = f_data, theta_k_val = theta_k_val, kpos = (k-1))
  if (gosolnp_result$values[length(gosolnp_result$values)] <= solnp_result$values[length(solnp_result$values)]) {
    return(gosolnp_result)
  } else {
    return(solnp_result)
  }
}

profileLL_values_theta0 <- rep(0,length(theta))
LRstat <- rep(Inf,length(theta))
CI_ptID_lower_bounds <- rep(-Inf,length(theta)) # Confidence interval lower bounds using conventional ML asymptotic normality.
CI_ptID_upper_bounds <- rep(Inf,length(theta)) # Confidence interval upper bounds using conventional ML asymptotic normality.
CI_CCT_M3_lower_bounds <- rep(-Inf,length(theta)) # Confidence interval lower bounds following Chen, Christensen, Tamer (2017) method 3 -- QLR.
CI_CCT_M3_upper_bounds <- rep(Inf,length(theta)) # Confidence interval upper bounds following Chen, Christensen, Tamer (2017) method 3 -- QLR.
Ln_max <- -Inf
theta_hat <- rep(0,length(theta))
std_errs <- rep(Inf,length(theta))

c_alpha <- qchisq(1-alpha,1) # critical value
tol <- 0.00001

# SEARCH FOR THE MAXIMIZER
  
  search.time.solnp <- system.time(
    solnp_LL_full <- solnp(par = theta, fun = neg.full.likelihood, LB = lower, UB = upper, control = list(trace = 0),
                           f_data = hhdata))
  cat("Full likelihood maximimzation by solnp took ", search.time.solnp[3], " seconds.\n", sep="")
  cat("Maximum value of the log likelihood was ", -solnp_LL_full$values[length(solnp_LL_full$values)], "\n", sep ="")
  
  search.time.gosolnp <- system.time(
    gosolnp_LL_full <- gosolnp(par = theta, fun = neg.full.likelihood, LB=lower, UB=upper, control = list(trace = 0),
                             n.sim = gosolnp_fullML_simulations, rseed = rseed, f_data = hhdata))
  cat("Full likelihood maximimzation by gosolnp using n.sim = ", gosolnp_fullML_simulations, " took ", search.time.gosolnp[3], " seconds.\n", sep="")
  cat("Maximum value of the log likelihood was ", -gosolnp_LL_full$values[length(gosolnp_LL_full$values)], "\n\n", sep ="")
  
  if (gosolnp_LL_full$values[length(gosolnp_LL_full$values)] < solnp_LL_full$values[length(solnp_LL_full$values)]) {
    LL_full_obj <- gosolnp_LL_full
  } else {
    LL_full_obj <- solnp_LL_full
  }
  Ln_max <- -LL_full_obj$values[length(LL_full_obj$values)]
  theta_hat <- LL_full_obj$pars
  hessian <- LL_full_obj$hessian
  
  if (is.positive.definite(hessian)) (std_errs <- sqrt(diag(chol2inv(chol(hessian)))))
  CI_ptID_lower_bounds <- theta_hat - qnorm(1-alpha/2) * std_errs
  CI_ptID_upper_bounds <- theta_hat + qnorm(1-alpha/2) * std_errs
  
  #################### Profile out theta[k] and compute CI ##########################################################
  for (k in parameter_indices) {
    
    # Starting value for full parameter vector in profile optimization searches.
    theta_start_k <- theta_hat
    
    # Set grid of theta[k] values in increments of stepsize[k]
    theta_k_values <- unique(c(rev(seq(theta_hat[k],lower[k],-stepsize[k])),seq(theta_hat[k],upper[k],stepsize[k])))
    
    # Profile LL values for each theta[k]
    theta_k_ll_values <- cbind(theta_k_values,-Inf)
    
    # Identify which of theta_k_values has kth component equal to that of the unconstrained theta_hat
    theta_hat_match_index <- which(theta_k_values==theta_hat[k])
    
    # Set the profile log likelihood for theta_hat[k] to be the uncontrainted maximum already computed
    theta_k_ll_values[theta_hat_match_index,2] <- Ln_max
    
    
    # form matrix to save optimal value of full parameter vector with theta[k] fixed at current value
    theta_hat_k_vec <- matrix(0, nrow = length(theta), ncol = length(theta_k_values))
    theta_hat_k_vec[,theta_hat_match_index] <- theta_hat
    
    cat("Computing profile log-likelihood for theta[", k, "] on a grid of ", length(theta_k_values)," values ranging from ", lower[k], " to ", upper[k], " in ", stepsize[k], " increments.\n", sep="")
    ############# Compute profile log-likelihood over the grid of values theta_k_values for theta[k] #####################################
    grid.search.time <- system.time (
    for (i in (1:length(theta_k_values))[-theta_hat_match_index]) { # profile LL known for theta_k[theta_hat_match_index]
      theta_start_k[k] <- theta_k_values[i]
      cat("Computing profile log likelihood for theta[", k, "] = ", theta_k_values[i], ".\n", sep="")
      search.time <- system.time (
        solnp_result <- k.profile.likelihood(theta_k_val = theta_k_values[i], theta_minus_k_start = theta_start_k[-k], 
                           LB_minus_k = lower[-k], UB_minus_k = upper[-k],
                           f_data = hhdata, k = k, nsim = gosolnp_profileML_simulations))
      
      theta_k_ll_values[i,2] <- -solnp_result$values[length(solnp_result$values)]
      theta_hat_k_vec[,i] <- append(solnp_result$pars, theta_k_values[i], after = (k-1))
      cat("Profile log likelihood maximimzation by solnp took ", search.time[3], " seconds.\n", sep="")
      cat("Profile log likelihood maximum value was ", theta_k_ll_values[i,2], ".\n\n", sep="")
    } )
    cat("Computing profile log-likelihood at each value on the grid took ", grid.search.time[3], " seconds.\n", sep="")
    Qn_vals_by_theta_k <- cbind(theta_k_values,2*(Ln_max - theta_k_ll_values[,2]))
  
    pass_null <- which(Qn_vals_by_theta_k[,2] <= c_alpha )
  
    LB_inner <- min(Qn_vals_by_theta_k[pass_null,1]) # minimum value of theta[k] on grid found to be less than the critical value
    UB_inner <- max(Qn_vals_by_theta_k[pass_null,1]) # maximum value of theta[k] on grid found to be less than the critical value
    
    LB_start <- theta_hat_k_vec[,min(pass_null)] # set starting value for LB search theta vector
    UB_start <- theta_hat_k_vec[,max(pass_null)] # set starting value for UB search theta vector
    
    LB_outer <- max(LB_inner - stepsize[k], lower[k])
    UB_outer <- min(UB_inner + stepsize[k], upper[k])
   
    cat("Searching for lower bound of CI for theta[", k, "].\n", sep="")
    start_time <- proc.time()[3]
    while ((LB_inner - LB_outer) > tol) {
      theta_k_try <- LB_inner - (LB_inner - LB_outer)/2
      search.time <- system.time (
        solnp_result <- k.profile.likelihood(theta_k_val = theta_k_try, theta_minus_k_start = LB_start[-k], 
                                             LB_minus_k = lower[-k], UB_minus_k = upper[-k],
                                             f_data = hhdata, k = k, nsim = gosolnp_search_simulations))
      profileLL_value <- -solnp_result$values[length(solnp_result$values)]
      LB_start <- append(solnp_result$pars, theta_k_try, after = (k-1)) # set starting value for next iteration
      Qnk_theta_try <- 2*(Ln_max - profileLL_value)
      ifelse (Qnk_theta_try <= c_alpha, LB_inner <- theta_k_try, LB_outer <- theta_k_try)
    }

    CI_CCT_M3_lower_bounds[k] <- LB_outer
    
    cat("Lower bound search took ", proc.time()[3] - start_time, " seconds.\n", sep="")
    cat("Lower bound = ", LB_outer, "\n", sep="")
    
    cat("Searching for upper bound of CI for theta[", k, "].\n", sep="")
    start_time <- proc.time()[3]
    while ((UB_outer - UB_inner) > tol) {
      theta_k_try <- UB_inner + (UB_outer - UB_inner)/2
      search.time <- system.time (
        solnp_result <- k.profile.likelihood(theta_k_val = theta_k_try, theta_minus_k_start = UB_start[-k], 
                                             LB_minus_k = lower[-k], UB_minus_k = upper[-k],
                                             f_data = hhdata, k = k, nsim = gosolnp_search_simulations))
      profileLL_value <- -solnp_result$values[length(solnp_result$values)]
      UB_start <- append(solnp_result$pars, theta_k_try, after = (k-1)) # set starting value for next iteration
      Qnk_theta_try <- 2*(Ln_max - profileLL_value)
      ifelse (Qnk_theta_try <= c_alpha, UB_inner <- theta_k_try, UB_outer <- theta_k_try)
    }
  
    CI_CCT_M3_upper_bounds[k] <- UB_outer
    cat("Upper bound search took ", proc.time()[3] - start_time, " seconds.\n", sep="")
    cat("Upper bound = ", UB_outer, "\n", sep="")
    
    cat("theta[", k, "] CI construction complete.\n", sep="")
    
  }
  
  cat("\n")
  cat("ML Estimate: ", theta_hat, "\n", sep="")
  cat("Wald CI lower endpoint: ", CI_ptID_lower_bounds, "\n", sep="")
  cat("Wald CI upper endpoint: ", CI_ptID_upper_bounds, "\n", sep="")
  
  for (k in parameter_indices) {
    cat("QLR CI lower endpoint for theta[", k, "] = ", CI_CCT_M3_lower_bounds[k], "\n", sep="")
    cat("QLR CI upper endpoint for theta[", k, "] = ", CI_CCT_M3_upper_bounds[k], "\n", sep="")
  }



