# Code for functions used in Aristodemou and Rosen (2021)

library(Rsolnp)

############# Function Compute_POR_LL_Linear computes the log-likelihood for the partially ordered
# response model with prices entering linearly.
# ---- Input parameters ---
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, rho, sigma2]
# B: N * 1 matrix of brand choices
# Y: N * 1 matrix of quality choices
# X1: N * k_beta1 matrix of covariates affecting brand 1 choice.
# X2: N * k_beta1 matrix of covariates affecting brand 2 choice.
#     Note X1 and X2 could contain common components or be identical.
# P: N * 2 * 2 array of prices. P[i,b,y] is the price for combination (b,y) faced by consumer i.
#     Note that ymax from a previous implementation is no longer used. The max quality choice is fixed at 2 in the current implementation.
# ---- Return parameters ---
# $likelihoods: N*1 vector of probabilities for each observed (B,Y).
# $iLL: N*1 vector of the log of the likelihoods above.
# $logLL: The log-likelihood, i.e. the sum of the components of iLL.
# $probs: N*1 vector of the probability of each observed choice computed by simulation.
# $simLL: The log-likelihood computed by simulation.
######################################################################################################
Compute_POR_LL_Linear <- function(theta,B,Y,X1,X2,P,weights=1) {

  k_gamma_delta <- 4
  k_beta1 <- dim(X1)[2]
  k_beta2 <- dim(X2)[2]
  N <- dim(X1)[1] # Sample Size
  
  rho <- theta[k_gamma_delta+k_beta1+k_beta2+1];
  if (abs(rho) > 1) { # rho needs to be a valid correlation coefficient. Return -Inf likelihood.
    return(list(likelihoods=rep(0,N),iLL=rep(-Inf,N),logLL=-Inf))
  }
  
  D <- 3-B # index of the brand not equal to B for each observation
  
  Xb <- (B==1) * X1 + (B==2) * X2 # covariates for chosen brand
  Xd <- (B==2) * X1 + (B==1) * X2 # covariates for unchosen brand
                                  # not used if outside option chosen.
  B1mat <- cbind(B==1,B==1)
  B2mat <- cbind(B==2,B==2)
  Pb <- B1mat * P[,1,] + B2mat * P[,2,]
  Pd <- B2mat * P[,1,] + B1mat * P[,2,]
  
  gamma <- theta[1:2] # There is a gamma parameter for each brand.
  delta <- theta[3:4] # There is a delta parameter for each brand.
  
  alpha <- array(0,dim = c(N,2,2)) # Store alpha for both firms. alpha[i,b,y] denotes alpha for brand b and quality y, observation i.
                             # In this specification alpha depends on price, and therefore varies across observations.
  alpha_1 <- delta[1] + gamma[1] * P[,1,]
  alpha_2 <- delta[2] + gamma[2] * P[,2,]
  alpha_b <- B1mat * alpha_1 + B2mat * alpha_2
  alpha_d <- B2mat * alpha_1 + B1mat * alpha_2
  
  # Create alpha_b and alpha_d with -Inf leading and Inf trailing to give cutoffs for
  # "Choice 0" and "Choice 3".
  alpha_b_prime <- cbind(-Inf,alpha_b,Inf)
  alpha_d_prime <- cbind(-Inf,alpha_d,Inf)
  alpha[,1,] = alpha_1
  alpha[,2,] = alpha_2
  
  beta <- matrix(0,nrow=max(k_beta1,k_beta2),ncol=2) # Store beta for both firms
  beta[1:k_beta1,1] <- theta[(k_gamma_delta+1):(k_gamma_delta+k_beta1)]
  beta[1:k_beta2,2] <- theta[(k_gamma_delta+k_beta1+1):(k_gamma_delta+k_beta1+k_beta2)]
  
  lambda_b <- cbind(pmin(alpha_b[,1],alpha_b[,2]/2), pmax(alpha_b[,2]-alpha_b[,1],alpha_b[,2]/2), Inf) - diag(Xb %*% beta[,B])
  sigma2 <- theta[k_gamma_delta+k_beta1+k_beta2+2];
  sigmavec <- c(1,sigma2) # vector of sigma1 and sigma2
  varmatrix <- matrix(c(1,rho*sigma2,rho*sigma2,sigma2^2),ncol=2,nrow=2)
  
  Yplus <- Y + (Y==0) # Set all zero entries to 1. This avoids an error indexing by zero,
                       # and is inconsequential because Y will not be used in this case.
  Yindices <- cbind(seq(1,N),Yplus) # matrix of indices to pick out prices for observed Y
  Zby <- ((delta[B] - delta[D] + gamma[B] * Pb[Yindices] + gamma[D]*(Pd[,2] - 2 * Pd[,1])) / Yplus) - diag(Xb %*% beta[,B])
  
  zero_cutoffs <- cbind(pmin(alpha_1[,1],alpha_1[,2]/2) - X1 %*% beta[,1], pmin(alpha_2[,1],alpha_2[,2]/2) - X2 %*% beta[,2])
  # Each column j of cutoffs gives the cutoff values for firm j for the outside good.
  
  p0 <- squish_in_place(pbivnorm(pmax(-100,pmin(100,zero_cutoffs[,1])),pmax(-100,pmin(100,zero_cutoffs[,2]/sigma2)),rho),0,1)
  # squish_in_place handles cases where the computational approximation to the bivariate normal results in pbivnorm producing
  # values outside the unit interval.  For example, it can produce probabilities of the order -1e-24 which should be zero.
  # calling squish_in_place is equivalent to using pmin and pmax as in the commented line below, but is faster.
  # p0 <- pmin(pmax(0,pbivnorm(pmax(-100,pmin(100,zero_cutoffs[,1])),pmax(-100,pmin(100,zero_cutoffs[,2]/sigma2)),rho)),1)
  
  sigma_B <- sigmavec[B]
  sigma_D <- sigmavec[D]
  
  Yplusone <- cbind(1:N,(Y+1))
  denom_plus <- sqrt( (sigma_B*Y)^2 - 4*rho*sigma_B*sigma_D*Y + 4*(sigma_D^2) )
  denom_minus <- sqrt( (sigma_B*Y)^2 - 2*rho*sigma_B*sigma_D*Y + sigma_D^2 )
  denom_matrix <- cbind(denom_plus,denom_plus,denom_minus,denom_minus)
  
  # set numerators for m1m2_matrix
  n1n2_matrix <- cbind( Y*diag(Xb %*% beta[,B]) + alpha[cbind(1:N,D,2)] - alpha_b_prime[Yplusone] - 2*diag(Xd %*% beta[,D]),
                        2*rho*sigma_D - sigma_B*Y,
                        Y*diag(Xb %*% beta[,B]) + alpha[cbind(1:N,D,1)] - alpha_b_prime[Yplusone] - diag(Xd %*% beta[,D]),
                        rho*sigma_D - sigma_B*Y )
  
  # Setting the m1m2 matrix deals with the possibility that the denom_matrix includes 0 elements, in which case 0/0 = NaN
  # Note: using squish_in_place is faster than using ifelse.
  m1m2_matrix <- 0*n1n2_matrix
  n1n2_nonzero <- which(n1n2_matrix!=0)
  m1m2_matrix[n1n2_nonzero] <- n1n2_matrix[n1n2_nonzero]/denom_matrix[n1n2_nonzero]
  
  if (any( abs(n1n2_matrix[,c(2,4)]) > abs(cbind(denom_plus,denom_minus)))) {
    # message indicates when truncating is used.
    # Investigations have indicated this only seems to happen when computational precision leads to machine values that
    # exceed one in absolute value by a tiny error due to approximation error. See comment on line 113.
    message("Max abs m1m2_matrix[,c(2,4)] is ", max(abs((m1m2_matrix[,c(2,4)]))), ".")
    message("Squishing m1m2_matrix[,c(2,4)] between -1 and 1.")
    m1m2_matrix[,c(2,4)] <- squish_in_place(m1m2_matrix[,c(2,4)],-1,1)
    # browser() # Used to investigate discrepancy in max(abs(n1n2_matrix[,c(2,4)])) - abs(cbind(denom_plus,denom_minus)))
      # This has always been found to be on the order of e-15, i.e. effectively zero. It appears to be sometimes infitesmally
      # greater than zero due to numerical imprecision.  Simple algebra shows it cannot be positive.
    } 
  
  lambda_b_y <- lambda_b[cbind(1:N,pmax(Y,1))]
  lambda_b_yplus1 <- lambda_b[cbind(1:N,Y+1)]
  
  whichY0 <- which(Y==0)
  whichT1 <- which((Y>0) & (Zby < lambda_b_yplus1))
  whichT2 <- which((Y>0) & (Zby > lambda_b_y))
  
  LLvec <- rep(0,N)
  LLvec[whichY0] <- p0[whichY0]
  
  LLvec[whichT1] <- Delta2(pmax(Zby[whichT1],lambda_b_y[whichT1])/(sigma_B[whichT1]),
                           lambda_b_yplus1[whichT1]/(sigma_B[whichT1]),
                           m1m2_matrix[whichT1,1],m1m2_matrix[whichT1,2])
  
  LLvec[whichT2] <- LLvec[whichT2] + Delta2(lambda_b_y[whichT2]/(sigma_B[whichT2]),
                                            pmin(Zby[whichT2],lambda_b_yplus1[whichT2])/(sigma_B[whichT2]),
                                            m1m2_matrix[whichT2,3],m1m2_matrix[whichT2,4]) 

  iLL = log(LLvec)
  logLL <- sum(iLL)
  
  return(list(likelihoods=LLvec,iLL=iLL,logLL=logLL))
  
}

Delta2 <- function(h,k,m1,m2) {
  return( (h < k) * squish_in_place(pbivnorm(pmax(-100,pmin(100,k)),pmax(-100,pmin(100,m1)),m2) - pbivnorm(pmax(-100,pmin(100,h)),pmax(-100,pmin(100,m1)),m2),0,1) )
}

pos.full.likelihood <- function(theta, f_data) {
  return(max(Compute_POR_LL_Linear(theta, f_data$B, f_data$Y, f_data$X1, f_data$X2, f_data$P)$logLL,-1000000000))
}

neg.full.likelihood <- function(theta, f_data) {
  return(min(-Compute_POR_LL_Linear(theta, f_data$B, f_data$Y, f_data$X1, f_data$X2, f_data$P)$logLL,1000000000))
}

# Log likelihood holding rho and sigma fixed, as specified in more_args.
# payoff_parameters
profiled.likelihood <- function(payoff_params, more_args, f_data) {
  theta <- c(payoff_params,more_args$rho,more_args$sigma)
  return(-Compute_POR_LL_Linear(theta, f_data$B, f_data$Y, f_data$X1, f_data$X2, f_data$P)$logLL)
}

# payoff_params: A starting value for the payoff paramters
rho.sigma.profiled.likelihood <- function(rho.sigma,payoff_params,f_data,payoff_params_LB,payoff_params_UB) {
  more_args <- list(rho = rho.sigma[1], sigma = rho.sigma[2])
  result <- solnp(par = payoff_params, fun = profiled.likelihood, LB = payoff_params_LB, UB = payoff_params_UB,
                  more_args = more_args, f_data = f_data, control = list(trace = 0))
  return(result$values[length(result$values)])
}


######## theta.k.profiled.likelihood #####################################################################
# Function theta.k.profiled.likelihood computes the profile likelihood for theta[k]
#####################################################################################################
theta.k.profiled.likelihood <- function(theta_k_val,theta_k_index,f_data,more_args) {
  theta_start <- more_args$theta_start
  theta_start[theta_k_index] <- theta_k_val
  LB <- more_args$LB
  LB[theta_k_index] <- theta_k_val
  UB <- more_args$UB
  UB[theta_k_index] <- theta_k_val
  result <- solnp(par = more_args$theta_start, fun = full.likelihood, LB = LB, UB = UB,
                  more_args = more_args, f_data = f_data, control = list(trace = 0))
  return(result$values[length(result$values)])
}

############# Function Sim_Prob computes conditional choice probabilities of all 5 possible choices
# by way of simulation for the partially ordered response model with prices entering linearly.
# ---- Input parameters ---
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, rho, sigma2]
# g: vector of length 2 such that g[j] represents firm j's gamma parameter
# d: vector of length 2 such that d[j] represents firm j's delta parameter
# b1: beta parameters for firm 1
# b2: beta parameters for firm2
# X1: vector of covariates affecting brand 1 choice.
# X2: vector of covariates affecting brand 2 choice.
#     Note X1 and X2 could contain common components or be identical.
# P: 2 * 2 array of prices. P[b,y] is the price for combination (b,y) faced by consumer i.
#     Note that ymax from a previous implementation is no longer used. The max quality choice is fixed at 2 in the current implementation.
# Udraws: N * 2 matrix representing R draws of the bivariate unobservables in the utility function.
######################################################################################################
Sim_Prob <- function(g,d,b1,b2,X1,X2,P,Udraws) {

  N <- dim(Udraws)[1]
  
  # Generate utilities from each (b,y) combination, for each observation.
  brand_utilities <- array(dim=c(N,2,2))
  brand_utilities[,1,1] <- c(X1 %*% b1) + Udraws[,1] - d[1] - g[1] * P[1,1]
  brand_utilities[,1,2] <- 2 * (c(X1 %*% b1) + Udraws[,1]) - d[1] - g[1] * P[1,2]
  brand_utilities[,2,1] <- c(X2 %*% b2) + Udraws[,2] - d[2] - g[2] * P[2,1]
  brand_utilities[,2,2] <- 2 * (c(X2 %*% b2) + Udraws[,2]) - d[2] - g[2] * P[2,2]
  
  brand1_utilities <- brand_utilities[,1,1:2]
  brand2_utilities <- brand_utilities[,2,1:2]
  max_brand1_utilities <- pmax(brand1_utilities[,1],brand1_utilities[,2])
  max_brand2_utilities <- pmax(brand2_utilities[,1],brand2_utilities[,2])
  max_overall_utilities <- pmax(max_brand1_utilities,max_brand2_utilities)
  B <- (max_brand2_utilities > max_brand1_utilities) + 1
  Y <- 1 * (pmax(brand1_utilities[,1],brand2_utilities[,1]) == max_overall_utilities) + 2 * (pmax(brand1_utilities[,2],brand2_utilities[,2]) == max_overall_utilities)
  Y <- Y * (max_overall_utilities > 0)
  
  p0 <- mean(Y==0)
  p11 <- mean((B==1) & (Y==1))
  p12 <- mean((B==1) & (Y==2))
  p21 <- mean((B==2) & (Y==1))
  p22 <- mean((B==2) & (Y==2))
  
  return(c(p0,p11,p12,p21,p22))
}

############# Function Compute_POR_ExpectedLL ###################################################
# computes the expected log-likelihood for the partially ordered
# response model with prices entering linearly.
# ---- Input parameters ---
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, rho, sigma2]
# Xsupport: |Support(X)| * length(X) matrix with each row representing a vector of possible realizations of X
#           The X variables are assumed to be individual specific and common to both brands.
# Psupport: |Support(P)| * 2 * 2 array with each 2*2 matrix in the last two dimensions representing prices
#           For each brand and quality. So Psupport[J,b,y] is the price for brand b and quality y at the
#           Jth support pont.
# XPprobs: |Support(X,P)| vector of marginal probabilities P(X=x) for x given by each row of Support (X)
# p0_BY_XP: Conditional probabilities of BY given XP in the population, each XP. |Support(X,P)| * 5 with
#           each row representing a differnt (X,P) and each column one of 0,(1,1),(1,2),(2,1),(2,2).
# ---- Return parameters ---
# $ELL: The expected log-likelihood
# $PYgivenXP: |Support(X,P)| * 5 vector of condtional probabilities for 0,(1,1),(1,2),(2,1),(2,2) given
#               the values of XP in the rows of XPsupport.
######################################################################################################
Compute_POR_ExpectedLL <- function(theta,Xsupport,Psupport,XPprobs,p0_BY_XP,cvar_obj) {
  theta_probs <- Compute_All_Probs(theta,Xsupport,Psupport,cvar_obj)
  tmp <- p0_BY_XP * log(theta_probs) * matrix(XPprobs, nrow = dim(p0_BY_XP)[1], ncol = 5)
  return(sum(tmp))
}

############# Function Compute_All_Probs #############################################################
# Compute conditional probabilities P(B,Y|X,P) for all X,P combinations on their respective supports
# ---- Input parameters ---
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, rho, sigma2]
# Xsupport: |Support(X)| * length(X) matrix with each row representing a vector of possible realizations of X
#           The X variables are assumed to be individual specific and common to both brands.
# Psupport: |Support(P)| * 2 * 2 array with each 2*2 matrix in the last two dimensions representing prices
#           For each brand and quality. So Psupport[J,b,y] is the price for brand b and quality y at the
#           Jth support pont.
# ---- Return parameters ---
# $PrBYgivenXP: |Support(X,P)| * 5 vector of condtional probabilities for 0,(1,1),(1,2),(2,1),(2,2) given
#               the values of XP in the rows of XPsupport.
######################################################################################################
Compute_All_Probs <- function(theta,Xsupport,Psupport,cvar_obj) {
  ### Unpack parameter vector ###
  k_gamma_delta <- 4
  k_beta1 <- dim(Xsupport)[2]
  k_beta2 <- k_beta1
  gamma <- theta[1:2] # There is a gamma parameter for each brand.
  delta <- theta[3:4] # There is a delta parameter for each brand.
  beta <- matrix(0, nrow = max(k_beta1,k_beta2), ncol = 2) # Store beta for both firms
  beta[1:k_beta1,1] <- theta[(k_gamma_delta+1):(k_gamma_delta+k_beta1)]
  beta[1:k_beta2,2] <- theta[(k_gamma_delta+k_beta1+1):(k_gamma_delta+k_beta1+k_beta2)]
  # Each column j of cutoffs gives the cutoff values for firm j
  rho <- theta[k_gamma_delta+k_beta1+k_beta2+1];
  sigma2 <- theta[k_gamma_delta+k_beta1+k_beta2+2];
  sigmavec <- c(1,sigma2) # vector of sigma1 and sigma2
  varmatrix <- matrix(c(1,rho*sigma2,rho*sigma2,sigma2^2),ncol=2,nrow=2)
  ######
  
  X1expand = cvar_obj$X1expand
  X2expand = cvar_obj$X2expand
  Pexpand = cvar_obj$Pexpand
  X1 = cvar_obj$X1
  X2 = cvar_obj$X2
  P = cvar_obj$P
  B = cvar_obj$B
  Y = cvar_obj$Y
  num_total_support_points = dim(X1)[1]
  
  # Use Compute_POR_LL_Linear to compute the conditional probabilities for each (B,Y) given (X,P)
  LL_result <- Compute_POR_LL_Linear(theta=theta,B=B,Y=Y,X1=X1expand,X2=X2expand,P=Pexpand,weights=1)
  
  # Extract the computed conditional probabilities in a num_total_support_points * 5 matrix
  PrBYgivenXP <- matrix(LL_result$likelihoods, nrow = num_total_support_points, ncol = 5)
  
  return(PrBYgivenXP)
}

############# Function Compute_CVar_Object #############################################################
# Compute covariate objects used in the computation of P(B,Y|X,P,theta) for all X,P combinations on their
# respective supports for a given value of theta.
# ******** FUNCTION IS NO LONGER USED *********
# ---- Input parameters ---
# Xsupport: |Support(X)| * length(X) matrix with each row representing a vector of possible realizations of X
#           The X variables are assumed to be individual specific and common to both brands.
# Psupport: |Support(P)| * 2 * 2 array with each 2*2 matrix in the last two dimensions representing prices
#           For each brand and quality. So Psupport[J,b,y] is the price for brand b and quality y at the
#           Jth support point.
######################################################################################################
Compute_CVar_Object <- function(Xsupport,Psupport) {
  
  nX <- dim(Xsupport)[1]
  nP <- dim(Psupport)[1]
  num_total_support_points <- nX * nP
  N <- num_total_support_points * 5 # Create 5 observations, one for each (B,Y) for each (X,P)
  X1 <- matrix(0, nrow = num_total_support_points, ncol = dim(Xsupport)[2])
  X2 <- matrix(0, nrow = num_total_support_points, ncol = dim(Xsupport)[2])
  P <- array(0, dim = c(num_total_support_points,2,2))
  j = 1
  cat("Constructing joint (X,P) support.\n")
  for (iX in 1:nX) {
    for (iP in 1:nP) {
      X1[j,] <- Xsupport[iX,]
      X2[j,] <- Xsupport[iX,]
      P[j,,] <- Psupport[iP,,]
      j <- j + 1 
    }
  }
  
  B <- rep(1,N)
  B[(num_total_support_points*3+1):N] <- 2
  Y <- rep(0,N)
  Y[(num_total_support_points+1):(num_total_support_points*2)] <- 1
  Y[(2*num_total_support_points+1):(num_total_support_points*3)] <- 2
  Y[(3*num_total_support_points+1):(num_total_support_points*4)] <- 1
  Y[(num_total_support_points*4+1):N] <- 2
  
  X1expand <- rbind(X1,X1,X1,X1,X1)
  X2expand <- rbind(X2,X2,X2,X2,X2)
  Pexpand <- array(0, dim = c(N,2,2))
  Pexpand[1:num_total_support_points,,] <- P
  Pexpand[(num_total_support_points+1):(2*num_total_support_points),,] <- P
  Pexpand[(2*num_total_support_points+1):(3*num_total_support_points),,] <- P
  Pexpand[(3*num_total_support_points+1):(4*num_total_support_points),,] <- P
  Pexpand[(4*num_total_support_points+1):(5*num_total_support_points),,] <- P
  
  ConditioningVarObjs <- list(X1expand = X1expand, X2expand = X2expand, Pexpand = Pexpand,
                              X1 = X1, X2 = X2, P = P, B = B, Y = Y)
  
}

############# Function squish_in_place #############################################################
# Produces the same output as pmin(pmax(a,x),b), returning a censored version of vector x where
# all values of x below a are set to a and all values above b are set to b.
# From the book Advanced R by Hadley Wickman.
# ---- Input parameters ---
# x: A numeric vector.
# a: A numeric scalar denoting the lower censoring point.
# b: A numeric scalre denoting the upper censoring point.
# ---- Return parameters ---
# pmin(pmax(a,x),b)
######################################################################################################
squish_in_place <- function(x,a,b) {
  x[x <= a] <- a
  x[x >= b] <- b
  return(x)
}

###############################################################################################################
############ CODE FOR COMPUTING ELASTICITIES FOLLOWS  ############################
###############################################################################################################

############# Function Compute_price_elasticities estimates the conditional price elasticity of demand
# of the given (b,y) alternative with respect to the price of the indicated alternative at the specified
# value of theta.  This is done by evaluating the given elasticity at theta for each value of observable
# exogenous variables in the data, and then taking the sample average of these elasticities.
# These are elasticities for alternative (bn,yn) with respect to the price of alternative (bd,yd).
#     'n' for numerator, 'd' for denominator
# ---- Input parameters ---
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, rho, sigma2]
# X1: N * k_beta1 matrix of covariates affecting brand 1 choice.
# X2: N * k_beta1 matrix of covariates affecting brand 2 choice.
#     Note X1 and X2 could contain common components or be identical.
# P: N * 2 * 2 array of prices. P[i,b,y] is the price for combination (b,y) faced by consumer i.
#     Note that ymax from a previous implementation is no longer used. The max quality choice is fixed at 2 in the current implementation.
# ---- Return parameters ---
# $POR_elasticities: elasticities of the share of each product (b,y) wrt price of each product (k,l).
# $outside_elasticities: elasticity of the share of the outside alternative wrt to price of each product (k,l)
######################################################################################################
Compute_price_elasticities <- function(theta,X1,X2,P) {
  
  # UNPACK INPUT PARAMETER AND DATA
  k_gamma_delta <- 4
  k_beta1 <- dim(X1)[2]
  k_beta2 <- dim(X2)[2]
  N <- dim(X1)[1] # Sample Size
  X <- array(0,c(N,k_beta1,2)) # In this version we are in fact assumed k_beta1 = k_beta2, but this can be easily generalized
  X[,,1] <- X1
  X[,,2] <- X2
  
  gamma <- theta[1:2] # There is a gamma parameter for each brand.
  delta <- theta[3:4] # There is a delta parameter for each brand.
    
  rho <- theta[k_gamma_delta+k_beta1+k_beta2+1];
  
  alpha <- array(0, dim = c(N,2,2)) # Store alpha for both firms. alpha[i,b,y] denotes alpha for brand b and quality y, observation i.
  # In this specification alpha depends on price, and therefore varies across observations.
  alpha[,1,] <- delta[1] + gamma[1] * P[,1,]
  alpha[,2,] <- delta[2] + gamma[2] * P[,2,]
  alpha_prime <- array(0, dim = c(N,2,4)) # Identical to alpha, but with -Inf leading and Inf trailing to give cutoffs for "Choice y=0" and "Choice y=3".
  alpha_prime[,,1] <- -Inf
  alpha_prime[,,2:3] <- alpha[,,]
  alpha_prime[,,4] <- Inf
  
  beta <- matrix(0,nrow=max(k_beta1,k_beta2),ncol=2) # Store beta for both firms
  beta[1:k_beta1,1] <- theta[(k_gamma_delta+1):(k_gamma_delta+k_beta1)]
  beta[1:k_beta2,2] <- theta[(k_gamma_delta+k_beta1+1):(k_gamma_delta+k_beta1+k_beta2)]
  
  sigma2 <- theta[k_gamma_delta+k_beta1+k_beta2+2]
  sigmavec <- c(1,sigma2) # vector of sigma1 and sigma2
  
  # For each b, lambda_cases[i,b] = T if inequality (4.4) holds, F otherwise.
  lambda_cases <- (alpha[,,2] > (2*alpha[,,1]))
  lambda <- array(0,c(N,2,3))
  # lambda for b=1
  lambda[,1,] <- cbind(pmin(alpha[,1,1],alpha[,1,2]/2), pmax(alpha[,1,2]-alpha[,1,1],alpha[,1,2]/2), Inf) - matrix(X1 %*% beta[,1],nrow=N,ncol=3)
  # lambda for b=2
  lambda[,2,] <- cbind(pmin(alpha[,2,1],alpha[,2,2]/2), pmax(alpha[,2,2]-alpha[,2,1],alpha[,2,2]/2), Inf) - matrix(X2 %*% beta[,2],nrow=N,ncol=3)
  
  # Set derivatives of lambda thresholds with respect to prices
  # dlambda_dp[i,b,y,k,l] is the derivative of lambda[i,b,y] with respect to price[k,l]
  dlambda_dp <- array(0,c(N,2,3,2,2))
  for (b in 1:2){
    dlambda_dp[,b,1,b,1] <- lambda_cases[,b] * gamma[b]
    dlambda_dp[,b,1,b,2] <- (1-lambda_cases[,b]) * (gamma[b]/2)
    dlambda_dp[,b,2,b,1] <- - lambda_cases[,b] * gamma[b]
    dlambda_dp[,b,2,b,2] <- lambda_cases[,b] * gamma[b] + (1-lambda_cases[,b]) * (gamma[b]/2)
  }
  
  # Set derivatives of zstar with respect to prices
  # dzstar_dp[b,y,k,l] is the derivative of zstar[b,y] with respect to price[k,l]
  dzstar_dp <- array(0,c(2,2,2,2))
  for (b in 1:2) {
    for (y in 1:2) {
      dzstar_dp[b,y,b,y] <- gamma[b] / y
      dzstar_dp[b,y,3-b,1] <- -2 * gamma[3-b] / y
      dzstar_dp[b,y,3-b,2] <- gamma[3-b] / y
    }
  }
  
  # Set derivatives m1_plus, m2_plus, m1_minus, m2_minus with respect to prices
  # First set denominator m1_plus, m2_plus, m1_minus, m2_minus and of derivatives
  # Each entry corresponds to the denominator for (b,y)
  denom_mplus <- array(0,c(2,2,2,2))
  denom_mplus[1,1,,] <- sqrt(sigmavec[1]^2 - 4*rho*sigmavec[1]*sigmavec[2] + (4*sigmavec[2]^2))
  denom_mplus[1,2,,] <- sqrt((4*sigmavec[1]^2) - 8*rho*sigmavec[1]*sigmavec[2] + (4*sigmavec[2]^2))
  denom_mplus[2,1,,] <- sqrt(sigmavec[2]^2 - 4*rho*sigmavec[1]*sigmavec[2] + (4*sigmavec[1]^2))
  denom_mplus[2,2,,] <- sqrt((4 * sigmavec[2]^2) - 8*rho*sigmavec[1]*sigmavec[2] + (4*sigmavec[1]^2))
  denom_mminus <- array(0,c(2,2,2,2))
  denom_mminus[1,1,,] <- sqrt(sigmavec[1]^2 - 2*rho*sigmavec[1]*sigmavec[2] + sigmavec[2]^2)
  denom_mminus[1,2,,] <- sqrt((4*sigmavec[1]^2) - 4*rho*sigmavec[1]*sigmavec[2] + sigmavec[2]^2)
  denom_mminus[2,1,,] <- sqrt(sigmavec[2]^2 - 2*rho*sigmavec[1]*sigmavec[2] + sigmavec[1]^2)
  denom_mminus[2,2,,] <- sqrt((4 * sigmavec[2]^2) - 4*rho*sigmavec[1]*sigmavec[2] + sigmavec[1]^2)
  
  # set numerators for m1m2_matrix and Zby
  # n1n2_array(i,j,b,y) is the numerator for observation i value of m1_plus, m2_plus, m1_minus, m2_minus
  # for brand b and quality y
  mplus <- array(0,c(N,2,2,2))
  mminus <- array(0,c(N,2,2,2))
  Zby <- array(0,c(N,2,2)) # this will be Zby_star.
  Zby_scaled <- array(0,c(N,2,2)) # this will be Zby_star divided by sigmavec[b].
  lambda_scaled <- array(0,c(N,2,3)) # this will be lambda divided by sigmavec[b].
  rep_sigma_vec <- array(0,c(N,2,2,2,2)) # Create array of repeated sigma values as need for later operations
  for (b in 1:2) {
    d <- 3-b
    rep_sigma_vec[,b,,,] <- sigmavec[b]
    lambda_scaled[,b,] <- lambda[,b,]/sigmavec[b]
    for (y in 1:2) {
      mplus[,,b,y] <- cbind(
        y*(X[,,b] %*% beta[,b]) + alpha[,d,2] - alpha_prime[,b,(y+1)] - 2*(X[,,d] %*% beta[,d]),
        2*rho*sigmavec[d] - sigmavec[b]*y) / denom_mplus[b,y,1,1]
      # note the last two entries of denom_mplus do not matter
      mminus[,,b,y] <- cbind(
        y*(X[,,b] %*% beta[,b]) + alpha[,d,1] - alpha_prime[,b,(y+1)] - (X[,,d] %*% beta[,d]),
        rho*sigmavec[d] - sigmavec[b]*y) / denom_mminus[b,y,1,1]
      # note the last two entries of denom_mminus do not matter
      Zby[,b,y] <- (( alpha[,d,2] + alpha[,b,y] - (2*alpha[,d,1]) ) / y) - (X[,,b] %*% beta[,b])
    }
    Zby_scaled[,b,] <- Zby[,b,]/sigmavec[b]
  }
  whichT1 <- (Zby < lambda[,,2:3])
  whichT2 <- (Zby > lambda[,,1:2])
  Zby_lambda_case <- 1*whichT1*whichT2 + 2*whichT2*(1-whichT1) + 3*whichT1*(1-whichT2)
  # For each b,y Zby_lambda_case[i,b,y] is 1 if lambda_b,y < Z_by < lambda_b,y+1,
  #                                        2 if lambda_b,y < Z_by and lambda_b,y+1 <= Z_by,
  #                                        3 if lambda_b,y >= Z_by and lambda_b,y+1 > Z_by,
  # This corresponds to the the three cases indicated in the proof of Proposition 2, (A.11) - (A.13)
  
  Delta_Partial_Terms <- array(0,c(N,2,2,4,4)) # Save partial derivatives of Delta function wrt to arguments
  Delta_Partial_Terms[,,,,1] <- Delta_Partials(lambda_scaled[,,1:2],Zby_scaled,mminus[,1,,],mminus[,2,,])
  Delta_Partial_Terms[,,,,2] <- Delta_Partials(Zby_scaled,lambda_scaled[,,2:3],mplus[,1,,],mplus[,2,,])
  Delta_Partial_Terms[,,,,3] <- Delta_Partials(lambda_scaled[,,1:2],lambda_scaled[,,2:3],mminus[,1,,],mminus[,2,,])
  Delta_Partial_Terms[,,,,4] <- Delta_Partials(lambda_scaled[,,1:2],lambda_scaled[,,2:3],mplus[,1,,],mplus[,2,,])
  
  # dmplus_dp[b,y,k,l] will be the derivative of m_plus[b,y] with respect to price[k,l]
  dmplus_dp <- array(0,c(2,2,2,2))
  dmplus_dp[1,1,2,2] <- gamma[2]
  dmplus_dp[1,1,1,1] <- -gamma[1]
  dmplus_dp[1,2,2,2] <- gamma[2]
  dmplus_dp[1,2,1,2] <- -gamma[1]
  dmplus_dp[2,1,1,2] <- gamma[1]
  dmplus_dp[2,1,2,1] <- -gamma[2]
  dmplus_dp[2,2,1,2] <- gamma[1]
  dmplus_dp[2,2,2,2] <- -gamma[2]
  dmplus_dp <- (dmplus_dp / denom_mplus)
  
  # dmminus_dp[b,y,k,l] will be the derivative of m_minus[b,y] with respect to price[k,l]
  dmminus_dp <- array(0,c(2,2,2,2))
  dmminus_dp[1,1,2,1] <- gamma[2]
  dmminus_dp[1,1,1,1] <- -gamma[1]
  dmminus_dp[1,2,2,1] <- gamma[2]
  dmminus_dp[1,2,1,2] <- -gamma[1]
  dmminus_dp[2,1,1,1] <- gamma[1]
  dmminus_dp[2,1,2,1] <- -gamma[2]
  dmminus_dp[2,2,1,1] <- gamma[1]
  dmminus_dp[2,2,2,2] <- -gamma[2]
  dmminus_dp <- (dmminus_dp / denom_mminus)
  
  # For each case, compute the partial derivative of the probability of (b,y) wrt to the price of product (k,l)
  # price_partials_case1 is an array of dimension Nx2x2x2x2.
  # For observations with lambda_b,y < Z_by < lambda_b,y+1, price_partials_case1[i,b,y,k,l] is
  # the partial derivative of Pr[B=b,Y=y|X=x_i W=W_i] with respect to W_(k,l), where W denotes price
  
  price_partials_case1 <-   (
    (( dlambda_dp[,,1:2,,] * array(Delta_Partial_Terms[,,,1,1],c(N,2,2,2,2))) / rep_sigma_vec
     + aperm(array(dmminus_dp,c(2,2,2,2,N)),c(5,1,2,3,4)) * array(Delta_Partial_Terms[,,,3,1],c(N,2,2,2,2)))
    + ((( dlambda_dp[,,2:3,,] * array(Delta_Partial_Terms[,,,2,2],c(N,2,2,2,2))) / rep_sigma_vec)
       + aperm(array(dmplus_dp,c(2,2,2,2,N)),c(5,1,2,3,4)) * array(Delta_Partial_Terms[,,,3,2],c(N,2,2,2,2))) )
  
  price_partials_case2 <-   ((( dlambda_dp[,,1:2,,] * array(Delta_Partial_Terms[,,,1,3],c(N,2,2,2,2))
                                + dlambda_dp[,,2:3,,] * array(Delta_Partial_Terms[,,,2,3],c(N,2,2,2,2))) / rep_sigma_vec)
                             + aperm(array(dmminus_dp,c(2,2,2,2,N)),c(5,1,2,3,4)) * array(Delta_Partial_Terms[,,,3,3],c(N,2,2,2,2)))
  
  price_partials_case3 <-   ((( dlambda_dp[,,1:2,,] * array(Delta_Partial_Terms[,,,1,4],c(N,2,2,2,2))
                                + dlambda_dp[,,2:3,,] * array(Delta_Partial_Terms[,,,2,4],c(N,2,2,2,2))) / rep_sigma_vec)
                             + aperm(array(dmplus_dp,c(2,2,2,2,N)),c(5,1,2,3,4)) * array(Delta_Partial_Terms[,,,3,4],c(N,2,2,2,2)))
   
  sby_price_partials <- array(0,c(N*2*2,2,2)) # partial derivatives of share of each (b,y) with respect to prices.
  pp_case_crunch_1 <- array(price_partials_case1,c(N*2*2,2,2))
  pp_case_crunch_2 <- array(price_partials_case2,c(N*2*2,2,2))
  pp_case_crunch_3 <- array(price_partials_case3,c(N*2*2,2,2))
  sby_price_partials[which(Zby_lambda_case==1),,] <- pp_case_crunch_1[which(Zby_lambda_case==1),,]
  sby_price_partials[which(Zby_lambda_case==2),,] <- pp_case_crunch_2[which(Zby_lambda_case==2),,]
  sby_price_partials[which(Zby_lambda_case==3),,] <- pp_case_crunch_3[which(Zby_lambda_case==3),,]
  
  # CONVERT OUTPUT FROM WHICH TO 3 INDICES (i,b,y)
  dim(sby_price_partials) <- c(N,2,2,2,2) # indices correspond to i,b,y,j,l
  
  s0_partials <- array(0,c(N,2,2))
  for(i in 1:N) {
    for (j in 1:2) {
      for (k in 1:2) {
        s0_partials[i,j,k] <- -sum(sby_price_partials[i,,,j,k]) # implied because probabilities must sum to one.
      }
    }
  }
  
  # Use Compute_POR_LL_Linear to compute the conditional probabilities for each (B,Y) given (X,P)
  # This is done by setting Y=0 for N obvservations across the different values of (X,P),
  # then setting B=1, Y=1 for N observations across all the different values of (X,P), and so on.
  LL_result00 <- Compute_POR_LL_Linear(theta=theta,B=c(rep(1,N)),Y=c(rep(0,N)),X1=X1,X2=X2,P=P)
  LL_result11 <- Compute_POR_LL_Linear(theta=theta,B=c(rep(1,N)),Y=c(rep(1,N)),X1=X1,X2=X2,P=P)
  LL_result12 <- Compute_POR_LL_Linear(theta=theta,B=c(rep(1,N)),Y=c(rep(2,N)),X1=X1,X2=X2,P=P)
  LL_result21 <- Compute_POR_LL_Linear(theta=theta,B=c(rep(2,N)),Y=c(rep(1,N)),X1=X1,X2=X2,P=P)
  LL_result22 <- Compute_POR_LL_Linear(theta=theta,B=c(rep(2,N)),Y=c(rep(2,N)),X1=X1,X2=X2,P=P)
  
  # Extract the computed conditional probabilities in an N * 5 matrix
  # The first column is the probability the outside good is chosen at each X_i, P_i
  # The second column is the probability (1,1) is chosen at each X_i, P_i
  # The third column is the probability (1,2) is chosen at each X_i, P_i, etc.
  PrBYgivenXP <- cbind(LL_result00$likelihoods,LL_result11$likelihoods,LL_result12$likelihoods,LL_result21$likelihoods,LL_result22$likelihoods)
  probs_0 <- PrBYgivenXP[,1]
  
  POR_elasticities <- array(0,c(N,2,2,2,2)) # 2nd dimension == 1 for Y=0.  2:(1,1), 3:(1,2), 4:(2,1), 5:(2,2)
  outside_elasticities <- s0_partials * (P / array(probs_0,c(N,2,2))) # elasticities for the outside good
  for (i in 1:2) {
    for (j in 1:2) {
      m <- 2*(i-1) + j + 1 # column index for Pr(B=i,Y=j) at each X_i,P_i in PrBYgivenXP
      for (k in 1:2) {
        for (l in 1:2) {
          zero_shares <- (PrBYgivenXP[,m]<1e-8) # market shares less than 1e-8 are set to zero.
          nz <- which(!zero_shares) # indices of entries with non-zero market share.
          # If the partial derivative is zero, set elasticity to zero to circumvent division by zero resulting in nan
          POR_elasticities[nz,i,j,k,l] <- sby_price_partials[nz,i,j,k,l] * (P[nz,k,l] / PrBYgivenXP[nz,m])  # implied because probabilities must sum to one.
          }
      }
    }
  }
  elasticities <- list(POR_elasticities = POR_elasticities, outside_elasticities = outside_elasticities)
  
  return(elasticities)
  
}

############# Function DeltaPartials #############################################################
# Computes the partial derivative of Delta2(h,k,m,rho) with respect to all arguments
# at the given value of input parameters.
# ---- Input parameters ---
# h,k,m,rho
# ---- Return parameters ---
# partials: A vector of partial derivatives with respect to each of the four arguments.
######################################################################################################
Delta_Partials <- function(h,k,m,rho) {
  a <- 1/sqrt(1-rho^2)
  partials <- cbind(- (exp(-(h^2)/2)/sqrt(2*pi)) * pnorm(a*(m-(h*rho))),
                    (exp(-(k^2)/2)/sqrt(2*pi)) * pnorm(a*(m-(k*rho))),
                    (exp(-(m^2)/2)/sqrt(2*pi)) * ( pnorm(a*(k-(m*rho))) - pnorm(a*(h-(m*rho))) ),
                    dnorm(k)*dnorm(a*(m - rho*k)) - dnorm(h)*dnorm(a*(m - rho*h)))
  return(partials)
}
