
####################### FUNCTION Execute_MC_Simulations #############################
# This functions executes a battery of Monte Carlos Simulations and writes the results
# to a file as specified.
#
# INPUT PARAMETERS: MCparameters 
Execute_MC_Simulations <- function(MCparameters) {
  
  # Unpack MCparameters
  R <- MCparameters$R
  theta <- MCparameters$theta
  lower <- MCparameters$lower
  upper <- MCparameters$upper
  stepsize <- MCparameters$stepsize
  alpha <- MCparameters$alpha
  gosolnp_full_ML_simulations <- MCparameters$gosolnp_full_ML_simulations
  gosolnp_profileML_simulations <- MCparameters$gosolnp_profileML_simulations
  parameter_indices <- MCparameters$parameter_indices
  
  c_alpha <- qchisq(1-alpha,1) # critical value
  tol <- 0.01
  
  for (DGP in MCparameters$DGPs) {
    for (N in MCparameters$SampleSizes) { # Run Monte Carlo for current sample size and DGP
    
      # Initialize
      simulated_MC_data <- Sim_POR_Data(N=N,theta=theta,R=R,DGP=DGP)
      profileLL_values_theta0 <- matrix(0, nrow = length(theta), ncol=R)
      LRstat <- matrix(Inf, nrow = length(theta), ncol = R)
      PDhessvec <- rep(0,R)
      PDhess.known.rho.sigma <- rep(0,R)
      CI_ptID_lower_bounds <- matrix(-Inf, nrow = length(theta), ncol = R) # Confidence interval lower bounds using conventional ML asymptotic normality.
      CI_ptID_upper_bounds <- matrix(Inf, nrow = length(theta), ncol = R) # Confidence interval upper bounds using conventional ML asymptotic normality.
      CI_ptID_lower_bounds_fixed_rho_sigma <- matrix(-Inf, nrow = (length(theta)-2), ncol = R) # Same as the above, but achieved by assuming that rho and sigma are known.
      CI_ptID_upper_bounds_fixed_rho_sigma <- matrix(Inf, nrow = (length(theta)-2), ncol = R)
      CI_CCT_M3_lower_bounds <- matrix(-Inf, nrow = length(theta), ncol = R) # Confidence interval lower bounds using Chen, Christensen, Tamer (2017) method 3.
      CI_CCT_M3_upper_bounds <- matrix(Inf, nrow = length(theta), ncol = R) # Confidence interval upper bounds using Chen, Christensen, Tamer (2017) method 3.
      Ln_max <- rep(-Inf,R)
      theta_hat <- matrix(0, nrow = length(theta), ncol = R)
      theta_1_14_hat <- matrix(0, nrow = (length(theta)-2), ncol = R) # ML estimates for the first 14 components of theta obtained fixing rho and sigma at their population values.
      std_errs <- matrix(Inf, nrow = length(theta), ncol = R)
      
      MC_times <- rep(0,R)
      
      #################
      for(r in 1:R) {
        iteration_start_time <- proc.time()[3]
        j <- (r-1) * N + 1 # Starting index of simulated draws for the r-th Monte Carlo sample
        k <- r * N # Ending index of simulated draws for the r-th Monte Carlo sample
        simdata <- list(B = simulated_MC_data$B[j:k], Y = simulated_MC_data$Y[j:k], X1 = simulated_MC_data$X1[j:k,], X2 = simulated_MC_data$X2[j:k,], P = simulated_MC_data$P[j:k,,])
        
        ######### MAXIMIZE THE LIKELIHOOD WITH RESPECT TO ALL PARAMETERS #######################################
        cat("Running Monte Carlo simulation ", r, " out of ", R, ".\n", sep="")
        search.time.solnp <- system.time(
          solnp_LL_full <- solnp(par = theta, fun = neg.full.likelihood, LB = lower, UB = upper, control = list(trace = 0),
                                 f_data = simdata))
        cat("Full likelihood maximimzation by solnp took ", search.time.solnp[3], " seconds.\n", sep="")
        cat("Maximum value of the log likelihood was ", -solnp_LL_full$values[length(solnp_LL_full$values)], "\n", sep ="")
        
        search.time.gosolnp <- system.time(
          gosolnp_LL_full <- gosolnp(par = theta, fun = neg.full.likelihood, LB=lower, UB=upper, control = list(trace = 0),
                                     n.sim = gosolnp_fullML_simulations, rseed = rseed, f_data = simdata))
        cat("Full likelihood maximimzation by gosolnp using n.sim = ", gosolnp_fullML_simulations, " took ", search.time.gosolnp[3], " seconds.\n", sep="")
        cat("Maximum value of the log likelihood was ", -gosolnp_LL_full$values[length(gosolnp_LL_full$values)], "\n\n", sep ="")
        
        if (gosolnp_LL_full$values[length(gosolnp_LL_full$values)] < solnp_LL_full$values[length(solnp_LL_full$values)]) {
          LL_full_obj <- gosolnp_LL_full
        } else {
          LL_full_obj <- solnp_LL_full
        }
        Ln_max[r] <- -LL_full_obj$values[length(LL_full_obj$values)]
        
        theta_hat[,r] <- LL_full_obj$pars
        hessian <- LL_full_obj$hessian
        
        if (is.positive.definite(hessian)) {
          tcresult <- tryCatch({
            std_errs[,r] <- sqrt(diag(chol2inv(chol(hessian))))
            PDhessvec[r] <- 1
            CI_ptID_lower_bounds[,r] <- theta_hat[,r] - qnorm(1-alpha/2) * std_errs[,r]
            CI_ptID_upper_bounds[,r] <- theta_hat[,r] + qnorm(1-alpha/2) * std_errs[,r]
          }, warning = function(w) {
          }, error = function(err) {
            PDhessvec[r] <- 0
            CI_ptID_lower_bounds[,r] <- Inf
            CI_ptID_upper_bounds[,r] <- -Inf
          }, finally = { }
          )
        } 
        ######### LIKELIHOOD MAXIMIZATION WITH RESPECT TO ALL PARAMETERS AND CI CONSTRUCTION COMPLETE #######################################
        
        ######### MAXIMIZE THE LIKELIHOOD WITH RESPECT TO ALL PAYOFF PARAMETERS, EVALUATED AT TRUE RHO, SIGMA #######################################
        solnp_LL_theta_1_14 <- solnp(par = LL_full_obj$pars[1:14], fun = p.payoff.likelihood, LB=lower[1:14], UB=upper[1:14], control = list(trace = 0), f_data = simdata)
        gosolnp_LL_theta_1_14 <- gosolnp(par = LL_full_obj$pars[1:14], fun = p.payoff.likelihood,
                                         LB=lower[1:14], UB=upper[1:14], control = list(trace = 0),
                                         n.sim = gosolnp_fullML_simulations, rseed = rseed, f_data = simdata)
        
        gosolnp_higher <- (gosolnp_LL_theta_1_14$values[length(gosolnp_LL_theta_1_14$values)] <= solnp_LL_theta_1_14$values[length(solnp_LL_theta_1_14$values)])
        if (gosolnp_higher) {
          theta_1_14_hat[,r] <- gosolnp_LL_theta_1_14$pars
          hessian <- gosolnp_LL_theta_1_14$hessian
        } else {
          theta_1_14_hat[,r] <- solnp_LL_theta_1_14$pars
          hessian <- solnp_LL_theta_1_14$hessian
        }
        
        PDhess.known.rho.sigma[r] <- is.positive.definite(hessian)
        if (PDhess.known.rho.sigma[r]) { # this will create a CI equal to the point estimate only if the hessian is not pd.
          known.rho.sigma.stderrs <- sqrt(diag(chol2inv(chol(hessian))))
        }
        else {
          known.rho.sigma.stderrs <- 0
        }
        
        CI_ptID_lower_bounds_fixed_rho_sigma[,r] <- theta_1_14_hat[,r] - qnorm(1-alpha/2) * known.rho.sigma.stderrs
        CI_ptID_upper_bounds_fixed_rho_sigma[,r] <- theta_1_14_hat[,r] + qnorm(1-alpha/2) * known.rho.sigma.stderrs
        
        ######### LIKELIHOOD MAXIMIZATION WITH RESPECT TO PAYOFF PARAMETERS AND CI CONSTRUCTION COMPLETE #######################################
        
        ######## Test true value of theta[k] for each k ####################################################################
        start_time <- proc.time()[3]
        for (k in parameter_indices) { 
          
          cat("Computing profile log likelihood for theta[", k, "].\n", sep="")
          search.time <- system.time (
            solnp_result <- k.profile.likelihood(theta_k_val = theta[k], theta_minus_k_start = theta[-k], 
                                                 LB_minus_k = lower[-k], UB_minus_k = upper[-k],
                                                 f_data = simdata, k = k, nsim = gosolnp_profileML_simulations)
            
            
          )
          
        
          cat("Profile log likelihood maximimzation by solnp took ", search.time[3], " seconds.\n", sep="")
          cat("Profile log likelihood maximum value was ", -solnp_result$values[length(solnp_result$values)], ".\n\n", sep="")
          
          profileLL_values_theta0[k,r] <- -solnp_result$values[length(solnp_result$values)]
          LRstat[k,r] <- 2*(Ln_max[r] - profileLL_values_theta0[k,r])
        }
        cat("Testing all parameter components took ", proc.time()[3] - start_time, " seconds.\n", sep="")
      
      MC_times[r] <- proc.time()[3] - iteration_start_time
      } # END MONTE CARLO LOOP ###############
      
      ## Set lower/upper bounds on fixed rho sigma CIs for sigma to be Inf/-Inf
      CI_ptID_lower_bounds_fixed_rho_sigma <- rbind(CI_ptID_lower_bounds_fixed_rho_sigma,Inf)
      CI_ptID_lower_bounds_fixed_rho_sigma <- rbind(CI_ptID_lower_bounds_fixed_rho_sigma,Inf)
      CI_ptID_upper_bounds_fixed_rho_sigma <- rbind(CI_ptID_upper_bounds_fixed_rho_sigma,-Inf)
      CI_ptID_upper_bounds_fixed_rho_sigma <- rbind(CI_ptID_upper_bounds_fixed_rho_sigma,-Inf)
      
      #########################
      ######### Record Results and Save Coverage Probabilities of Monte Carlos ###########
      ###### DISPLAY OUTPUT ###############
      sink(paste("POR_MCs_Tests_", Sys.time(), ", N = ", N, " DGP = ", DGP, ".txt", sep=""))
      cat(" ---- Partially Ordered Response Monte Carlo Coverage Probabilities --- \n", sep="")
      cat("Sample size N = ", N, "\n", sep = "" )
      cat("Number of Monte Carlo iterations executed = ", R, "\n", sep = "")
      cat("DGP = ", DGP, ".\n", sep="")
      cat("gosolnp_fullML_simulations = ", gosolnp_fullML_simulations, ".\n", sep="")
      cat("gosolnp_profileML_simulations = ", gosolnp_profileML_simulations, ".\n", sep="")
      cat("Nominal coverage probabilities = ", 1-alpha, "\n", sep = "")
      cat("Parameter vector theta = ", capture.output(theta), "\n", sep = "")
      cat("Feasible ML hessian was positive definite ", sum(PDhessvec), " times.\n", sep="")
      cat("Infeasible ML hessian was positive definite ", sum(PDhess.known.rho.sigma), " times.\n", sep="")
      coverage_probs <- matrix(0, nrow = length(theta), ncol = 3)
      for (k in parameter_indices) {
        coverage_probs[k,1] <- mean((CI_ptID_lower_bounds_fixed_rho_sigma[k,] <= theta[k]) & (theta[k] <= CI_ptID_upper_bounds_fixed_rho_sigma[k,]))
        coverage_probs[k,2] <- mean((CI_ptID_lower_bounds[k,] <= theta[k]) & (theta[k] <= CI_ptID_upper_bounds[k,]))
        coverage_probs[k,3] <- mean(LRstat[k,] <= c_alpha)
        cat("Coverage probabilities for parameter ", k, ": ",
            coverage_probs[k,1], ", ", coverage_probs[k,2], ", ", coverage_probs[k,3], "\n", sep = "")
      }
      
      
      sink()
      #############################################
    }
  }
  return(MC_times)

}