
library("mvtnorm")

############# Function Sim_POR_Data generate data for the partially ordered response model with prices entering linearly
# ---- Input parameters ---
# N: The number of observations per sample
# theta: the vector of payoff and distributional parameters [gamma1, gamma2, delta1, delta2, beta1, beta2, sigma2, rho]
# R: The number of samples to generate. The default is 1.
# DGP: The index of the DGP from which to sample. The default is 1.
# ---- Return parameters ---
# The following are elements of a list that is returned
# B: N * 1 matrix of brand choices
# Y: N * 1 matrix of quality choices
# X1: N * k_beta1 matrix of covariates affecting brand 1 choice.
# X2: N * k_beta1 matrix of covariates affecting brand 2 choice.
#     Note that here k_beta1 and k_beta2 must be 5.
#     Note X1 and X2 could contain common components or be identical.
# P: N * 2 * 2 array of prices. P(i,b,y) is the price for combination (b,y) faced by consumer i.
#     Note that ymax from a previous implementation is no longer used. The max quality choice is fixed at 2 in the current implementation.
######################################################################################################
Sim_POR_Data <- function(N,theta,R=1,DGP=1) {
  
  # Unpack components of theta
  gamma <- theta[1:2] # There is a gamma parameter for each brand.
  delta <- theta[3:4]
  beta1 <- theta[5:9]
  beta2 <- theta[10:14]
  rho <- theta[15]
  sigma <- theta[16]
  
  # Generate regressors
  age_var <- runif(N*R,0,100)
  age_31_40 <- (age_var<42.6) # This will generate probabilities of similar magnitude to what is in the data
  age_41_50 <- (age_var>76.6)
  married <- rbinom(N*R,1,0.4)
  employed <- rbinom(N*R,1,0.85)
  more_females <- (runif(N*R,0,100) < 55.4)
  X1 <- cbind(age_31_40,age_41_50,married,employed,more_females)
  X2 <- X1
  
  # Generate prices
  price_rho <- 0.25 # correlation for temporary parameter in generation of prices
  prices <- array(dim = c(N*R,2,2))
  price_var_matrix <- matrix(c(1,price_rho,price_rho,1),nrow = 2)
  
  if (DGP == 1) { # Continuous, unbounded prices with Y=1 chosen with zero probability for some prices
    prices[,1,1] <- runif(N*R,1,4)
    prices[,2,1] <- runif(N*R,1.35,2.15)
    tmp <- rmvnorm(N*R, mean = c(0,0), sigma = price_var_matrix)
    prices[,1,2] <- prices[,1,1] + tmp[,1]
    prices[,2,2] <- prices[,2,1] + tmp[,2]
  } else if (DGP == 2) { # Continuous, bounded prices.
    prices[,1,1] <- runif(N*R,1,2)
    prices[,2,1] <- runif(N*R,1.35,2.15)
    tmp <- abs(rmvnorm(N*R, mean = c(0,0), sigma = price_var_matrix))
    prices[,1,2] <- prices[,1,1] + pmax(1,pmin(2,tmp[,1]))
    prices[,2,2] <- prices[,2,1] + pmax(1,pmin(2,tmp[,2]))
  } else { # Prices with discrete support
    prices[,1,1] <- round(runif(N*R,1,2))
    prices[,2,1] <- round(runif(N*R,1.35,2.15))
    tmp <- abs(rmvnorm(N*R, mean = c(0,0), sigma = price_var_matrix))
    prices[,1,2] <- prices[,1,1] + round(pmax(1,pmin(2,tmp[,1])))
    prices[,2,2] <- prices[,2,1] + round(pmax(1,pmin(2,tmp[,2])))
  } 
  
  # Generate utility function unobservables
  U_var_matrix <- matrix(c(1,rho*sigma,rho*sigma,sigma^2),nrow = 2)
  U <- rmvnorm(N*R, mean = c(0,0), sigma = U_var_matrix)
  
  # Generate utilities from each (b,y) combination, for each observation.
  brand_utilities <- array(dim=c(N*R,2,2))
  brand_utilities[,1,1] <- X1 %*% beta1 + U[,1] - delta[1] - gamma[1] * prices[,1,1]
  brand_utilities[,1,2] <- 2 * (X1 %*% beta1 + U[,1]) - delta[1] - gamma[1] * prices[,1,2]
  brand_utilities[,2,1] <- X2 %*% beta2 + U[,2] - delta[2] - gamma[2] * prices[,2,1]
  brand_utilities[,2,2] <- 2 * (X2 %*% beta2 + U[,2]) - delta[2] - gamma[2] * prices[,2,2]

  brand1_utilities <- brand_utilities[,1,1:2]
  brand2_utilities <- brand_utilities[,2,1:2]
  max_brand1_utilities <- pmax(brand1_utilities[,1],brand1_utilities[,2])
  max_brand2_utilities <- pmax(brand2_utilities[,1],brand2_utilities[,2])
  max_overall_utilities <- pmax(max_brand1_utilities,max_brand2_utilities)
  B <- (max_brand2_utilities > max_brand1_utilities) + 1
  Y <- 1 * (pmax(brand1_utilities[,1],brand2_utilities[,1]) == max_overall_utilities) + 2 * (pmax(brand1_utilities[,2],brand2_utilities[,2]) == max_overall_utilities)
  Y <- Y * (max_overall_utilities > 0)
  
  return(list(B = B, Y = Y, X1 = X1, X2 = X2, P = prices)) 
}